#ifndef DEFS_HPP
#define DEFS_HPP
#
#/* Copyright (C) 2003  Vesa Karvonen
# *
# * Permission to copy, use, modify, sell and distribute this software is
# * granted provided this copyright notice appears in all copies. This
# * software is provided "as is" without express or implied warranty, and with
# * no claim as to its suitability for any purpose.
# */
#
#/* Definitions for some lambda-terms */
#
#/* Some syntactic sugar */
#define DELAY(e) LAMBDA(_,e)
#define FORCE(e) (e,ID)
#define LET(s,i,t) (LAMBDA(s,t),i)
#define IF(c,t,e) FORCE(((c,DELAY(t)),DELAY(e)))
#
#/* Identity */
#define ID LAMBDA(X,X)
#
#/* Church-booleans */
#define TRUE LAMBDA(T,LAMBDA(_,T))
#define FALSE LAMBDA(_,LAMBDA(F,F))
#define AND LAMBDA(A,LAMBDA(B,((A,B),FALSE)))
#define TEST LAMBDA(L,LAMBDA(M,LAMBDA(N,((L,M),N))))
#
#/* Pairs */
#define PAIR LAMBDA(F,LAMBDA(S,LAMBDA(B,((B,F),S))))
#define FST LAMBDA(P,(P,TRUE))
#define SND LAMBDA(P,(P,FALSE))
#
#/* Church-numerals */
#define ZERO LAMBDA(_,LAMBDA(Z,Z))
#define ONE LAMBDA(S,LAMBDA(Z,(S,Z)))
#define TWO LAMBDA(S,LAMBDA(Z,(S,(S,Z))))
#define THREE LAMBDA(S,LAMBDA(Z,(S,(S,(S,Z)))))
#define SUCC LAMBDA(N,LAMBDA(S,LAMBDA(Z,(S,((N,S),Z)))))
#define ZZ LAMBDA(B,((B,ZERO),ZERO))
#define SS LAMBDA(P,LET(F,(P,FALSE),LET(S,(SUCC,F),LAMBDA(B,((B,F),S)))))
#define PRED LAMBDA(M,(((M,SS),ZZ),TRUE))
#define PLUS LAMBDA(M,LAMBDA(N,LAMBDA(S,LAMBDA(Z,((M,S),((N,S),Z))))))
#define TIMES LAMBDA(M,LAMBDA(N,((M,(PLUS,N)),ZERO)))
#define IS_ZERO LAMBDA(M,((M,LAMBDA(_,FALSE)),TRUE))
#
#/* Fixed-point combinator */
#define FIX LAMBDA(F,LET(G,LAMBDA(X,(F,LAMBDA(Y,((X,X),Y)))),(G,G)))
#
#/* Factorial */
#define FACT (FIX,LAMBDA(F,LAMBDA(N,LET(M,(PRED,N),IF((IS_ZERO,M),ONE,((TIMES,N),(F,M)))))))
#
#endif
